import os
import requests
import mimetypes

def get_headers(token):
    return {
        'Authorization': f'Bearer {token}',
        'product': 'niopub',
    }

class NeosphereMediaClient:
    def __init__(self, token, media_directory, url):
        """
        Initialize the NeosphereMediaClient with an API token and a media directory.

        Parameters:
            token (str): The API token for authentication.
            media_directory (str): The directory where media files will be saved.

        The constructor ensures that the media_directory exists or creates it.
        """
        self.token = token
        # the base https url
        self.base_url = url+"media/"
        self.headers = get_headers(token)
        self.media_directory = media_directory

        # Ensure the media_directory exists or create it
        if not os.path.exists(self.media_directory):
            os.makedirs(self.media_directory, exist_ok=True)
        elif not os.path.isdir(self.media_directory):
            raise NotADirectoryError(f"The path '{self.media_directory}' is not a directory.")
    
    def create_forward_copy_id(self, forward_to_id, media_id):
        """
        Create a copy of the media for a new recipient.
        Prevents the need to download and re-upload the media.

        Parameters:
            media_id (str): The ID of the media to copy.
            forward_to_id (str): The ID of the recipient group or agent-id.

        Returns:
            dict: The JSON response from the API for the new media.
        """
        # Fetch the media data using the media_id.
        # Currently we only support with param ?for_agent=true
        url = self.base_url+f"forward/{forward_to_id}/{media_id}?for_agent=true"
        response = requests.post(url, headers=self.headers, stream=True)
        response.raise_for_status()
        # get the media ID from response json
        media_id = response.json()['media_id']
        return media_id

    def get_media(self, media_id)->str:
        """
        Retrieve media data by media_id and save it to the media_directory.

        Parameters:
            media_id (str): The ID of the media to retrieve.

        Returns:
            str: The file path where the media is saved.
        """
        file_path = ""
        url = f"{self.base_url}{media_id}"
        response = requests.get(url, headers=self.headers, stream=True)
        response.raise_for_status()

        # Determine the filename
        # Try to get filename from 'Content-Disposition' header
        content_disposition = response.headers.get('Content-Disposition')
        content_type = response.headers.get('Content-Type')
        if content_disposition:
            # Parse filename from content_disposition
            import re
            filename_match = re.search(r'filename="?([^"]+)"?', content_disposition)
            if filename_match:
                filename = filename_match.group(1)
            else:
                filename = f"{media_id}"
        else:
            # Fallback to media_id as filename
            filename = f"{media_id}" + "." + content_type.split('/')[-1]

        # Save the file to media_directory
        file_path = os.path.join(self.media_directory, filename)
        print(file_path)

        # Write the content to the file
        with open(file_path, 'wb') as f:
            for chunk in response.iter_content(chunk_size=8192):
                if chunk:
                    f.write(chunk)

        return file_path

    def save_media(self, parent_id, media_file, filename=None, content_type=None):
        """
        Save or update media data by attached resource ID and it's file-like object.

        Parameters:
            parent_id (str): The ID of the resource to attach the media to, can be the group ID the agent is
                responding to or another agent's share_id if sending the media to another online agent.
            media_id (str): The ID of the media to save.
            media_file (file-like object): The file-like object containing media data.
            filename (str, optional): The filename to use. If not provided, it will be
                                    determined from media_file's deduced name.
            content_type (str, optional): The MIME type of the media. If not provided,
                                          it will be guessed based on the filename.

        Returns:
            str: The new Media ID created.
        """
        url = f"{self.base_url}/{parent_id}"

        # If filename is not provided, try to get it from the file-like object
        if filename is None:
            filename = getattr(media_file, 'name', None)
            if filename is None:
                raise ValueError("Filename must be provided if media_file has no 'name' attribute.")

            # Get only the basename if media_file.name is a path
            filename = os.path.basename(filename)

        # If content_type is not provided, guess it based on the filename
        if content_type is None:
            content_type = mimetypes.guess_type(filename)[0] or 'application/octet-stream'

        files = {
            'file': (filename, media_file, content_type)
        }

        response = requests.put(url, headers=self.headers, files=files)
        response.raise_for_status()
        media_id = response.json()['media_id']
        return media_id
    
    def upload_media_from_path(self, media_id, file_path):
        """
        Helper method to upload media from an absolute file path.

        Parameters:
            media_id (str): The ID of the media to upload.
            file_path (str): The absolute path to the file.

        Returns:
            dict: The JSON response from the API.
        """
        if not os.path.isabs(file_path):
            raise ValueError("The file path must be an absolute path.")

        if not os.path.exists(file_path):
            raise FileNotFoundError(f"The file '{file_path}' does not exist.")

        filename = os.path.basename(file_path)
        content_type = mimetypes.guess_type(filename)[0] or 'application/octet-stream'

        with open(file_path, 'rb') as media_file:
            return self.save_media(media_id, media_file, filename=filename, content_type=content_type)


import asyncio
import logging
import json
from typing import Any, Optional

from websockets import connect as ws_connect
from websockets.client import WebSocketClientProtocol
from websockets.exceptions import ConnectionClosedError, ConnectionClosedOK, InvalidStatus
from neosphere.client_api import Message, NeosphereClient
from neosphere.contacts_handler import NeosphereAgentContactsClient
from neosphere.media_handler import NeosphereMediaClient
logger = logging.getLogger('neosphere').getChild(__name__)
import traceback

NEOSPHERE_DNS_NAME = "n10s.net"
class NeosphereAgent:
    """
    A client agent for the Neosphere network that manages WebSocket connections,
    message handling, and callback triggering for both human and AI messages.
    
    This class abstracts the logic required to:
    - Connect and authorize with the Neosphere server.
    - Process incoming messages and dispatch them via provided callbacks.
    - Handle reconnections and error states gracefully.
    
    Parameters:
        agent_share_id: Unique identifier for the agent.
        connection_code: Authentication code for establishing connection.
        client_name: The display name for the client.
        group_message_receiver: Callback to process group messages from human users.
        query_receiver: Callback to process queries from other AI agents.
        contacts: Initial contacts to use for message handling.
        **context: Additional context passed to callbacks.
    """
    app: Any
    ws: WebSocketClientProtocol

    def __init__(self, 
                 agent_share_id, 
                 connection_code, 
                 client_name, 
                 group_message_receiver, 
                 query_receiver,
                 contacts=[],
                 **context) -> None:
        self.agent_share_id = agent_share_id
        self.connection_code = connection_code
        self.client_name = client_name
        self.human_group_msg_callback = group_message_receiver
        self.ai_query_msg_callback = query_receiver
        self.reconnection_token = None
        self.context_to_forward = context
        self.http_url = None
        # query tracker is any simple key value structure the client will use to keep a track of queries sent and their responses received.
        # this allows the client to provide a blocking wait for a query response.
        self.query_tracker = {}
        self.neosphere_client = NeosphereClient(query_index=self.query_tracker, name=agent_share_id)
        self.recieved_pull_the_plug = False
        self.initial_contacts = contacts


    async def set_websocket(self, ws: WebSocketClientProtocol):
        self.ws = ws

    async def before_connect(self):
        pass

    async def attempt_reconnect(self):
        pass

    async def on_connect(self):
        pass
    
    async def on_authorize(self, ignore_token=False):
        if self.reconnection_token and not ignore_token:
            logger.info("Restarting agent connection with connection token")
            await self.neosphere_client.send_token_to_reconnect(self.reconnection_token, self.agent_share_id)
        else:
            logger.info("Authorizing agent connection with connection code")
            await self.neosphere_client.send_token_request(self.connection_code, self.agent_share_id, self.client_name)
    
    async def before_disconnect(self):
        logger.info('before_disconnect')

    async def on_disconnect(self):
        logger.info('on_disconnect')
    
    def set_http_requests_address(self, http_url: str):
        self.http_url = http_url
    
    async def on_message(self, message: str):
        msg = Message.from_json(message)
        logger.debug(f"Received processed message: {msg.to_dict()}")
        try:
            if msg.is_pull_the_plug():
                logger.info("Received pull the plug signal. Closing connection.")
                self.recieved_pull_the_plug = True
                await self.neosphere_client.send(None)
            if msg.token:
                self.reconnection_token = msg.token
                self.neosphere_client.register_media_handler(NeosphereMediaClient(self.reconnection_token, "/tmp/neosphere_media", self.http_url))
                self.neosphere_client.register_contacts_handler(NeosphereAgentContactsClient(self.reconnection_token, self.initial_contacts, self.http_url))
                # initialize contacts
                # Contacts().initial_public_contacts(public_contacts)
                logger.info(f"Received a connection token. Registered media handler and fetched ({self.neosphere_client.contacts.get_contact_count()}) contacts.")
            if msg.is_err:
                logger.error(f"Received error message.")
            if msg.group_id:
                if not msg.is_err:
                    logger.info('Message is group message')
                    await self.human_group_msg_callback(msg, self.neosphere_client, **self.context_to_forward)
                else:
                    logger.error(f"(Group={msg.group_id}) Error from: {msg.from_id}. Error: {msg.text}")
            elif msg.query_id:
                if msg.is_err:
                    if msg.from_id == "sys":
                        logger.error(f"Message is a Neosphere system error on query ID: {msg.query_id} - {msg.text}. Will not record this response as it is a system error, will let the caller timeout.")
                    else:
                        logger.error(f"Message is an error on query ID: {msg.query_id}")
                        await self.neosphere_client._record_query_response_recvd(msg.query_id, msg)
                elif msg.is_resp:
                    logger.info('Message is a query resp')
                    await self.neosphere_client._record_query_response_recvd(msg.query_id, msg)
                else:
                    logger.info('Message is a query from another agent')
                    await self.ai_query_msg_callback(msg, self.neosphere_client, **self.context_to_forward)
        except Exception as e:
            logger.error(f"Error while processing message: {e}")
            traceback.print_exc()
            return


class NeosphereAgentTaskRunner(object):
    """
    Manages the lifecycle of a NeosphereAgent's WebSocket connection.
    
    This class handles:
    - Establishing and cleaning up Neosphere connections.
    - Authorizing the agent after connecting.
    - Running separate asynchronous loops for sending and receiving messages.
    - Retrying connection attempts up to a defined maximum.
    
    Parameters:
        agent: An instance of NeosphereAgent to run.
        url: Optional URL for the Neosphere server.
    """
    
    hostname: str
    agent: NeosphereAgent
    ws: WebSocketClientProtocol

    # states
    _is_connected: bool
    _authorize_called: bool
    _listening: bool
    _auth_with_conn_code: bool
    _retry_count: int
    _err_state: bool
    MAX_RETRIES = 50

    def __init__(self, agent: NeosphereAgent, server_hostname: str = None) -> None:
        self.agent = agent
        self.hostname = server_hostname or NEOSPHERE_DNS_NAME
        self.agent.set_http_requests_address(self._get_url(ws=False))
        self._authorize_called = False
        self._listening = False
        self._is_connected = False
        self._auth_with_conn_code = True
        self._retry_count = 0
        self._err_state = False
        self.recv_active = []
        self.send_active = []
        self.GAP_BETWEEN_RETRIES_SEC = 2

    def slowdown_retries(self):
        self.GAP_BETWEEN_RETRIES_SEC = 10
    def reset_retry_speed(self):
        self.GAP_BETWEEN_RETRIES_SEC = 2
    
    def _get_url(self, ws=True):
            if self.hostname.startswith("localhost"):
                if ws:
                    proto = "ws://"
                else:
                    proto = "http://"
            else:
                if ws:
                    proto = "wss://"
                else:
                    proto = "https://"
            return f"{proto}{self.hostname}/"

    def _clean_state(self):
        self._is_connected = False
        self._listening = False
        self._err_state = False

    async def connect(self):
        self._clean_state()
        self._err_state = False
        await self.agent.before_connect()
        try:
            self.ws = await ws_connect(self._get_url()+"stream/ai")
        except ConnectionRefusedError as e:
            logger.error(f'Connection refused: {e}')
            return
        except asyncio.TimeoutError as e:
            logger.error(f'Timeout when connecting: {e}')
            return
        except InvalidStatus as e:
            logger.error(f'Invalid status: {e}')
            self.slowdown_retries()
            return
        await self.agent.set_websocket(self.ws)
        await self.agent.on_connect()
        self.reset_retry_speed()
        self._is_connected = True
    
    async def authorize(self):
        if not self._is_connected:
            logger.error("Cannot authorize as not connected")
            return
        await self.agent.on_authorize(ignore_token=self._auth_with_conn_code)
        # this only gets set from the recv_loop, reset it once it's used
        self._auth_with_conn_code = False
        self._authorize_called = True

    async def disconnect(self):
        self._is_connected = False
        await self.agent.before_disconnect()
        await self.ws.close()
        await self.agent.on_disconnect()

    async def send(self, message: str) -> Any:
        await self.ws.send(message)

    async def ws_recv_message(self) -> Optional[str]:
        try:
            return await asyncio.wait_for(self.ws.recv(), 1)

        except asyncio.TimeoutError:
            return None

    async def ws_recv_loop(self):
        while self._is_connected:
            try:
                if not self._listening:
                    self._listening = True 
                message = await self.ws.recv()
            except ConnectionClosedError as e:
                logger.error(f'Recieve loop closed with error: {e.reason}, code: {e.code}')
                if e.code == 4010:
                    # this will cause the agent to reauthorize with the connection code
                    # useful when code is pinned, auth will still fail if code also expired on the app
                    self._auth_with_conn_code = True
                self._err_state = True
                # sending None triggers an iteration in the send loop
                await self.agent.neosphere_client.send(None)
                break
            except ConnectionClosedOK as e:
                logger.info(f'Recieve loop closed without error: {e}')
                # this will prevent retrying to open the connection.
                self._retry_count = self.MAX_RETRIES + 1
                self._is_connected = False
                # sending None triggers an iteration in the send loop
                await self.agent.neosphere_client.send(None)
                break
            if message is None:
                continue

            # call on_message async function as a task and keep moving
            # await self.reciever.on_message(message)
            task = asyncio.create_task(self.agent.on_message(message))
            self.recv_active.append(task)
        logger.info(f"Recieve loop ending. Will wait for all incoming messages ({len(self.recv_active)}) to finish.")
        if len(self.recv_active) > 0:
            await asyncio.gather(*self.recv_active)
        logger.info(f"Recieve loop ended.")

    async def ws_send_loop(self):
        while self._is_connected:
            if not self._listening:
                # sleep till listening is enabled
                await asyncio.sleep(0.2)
                continue
            if self._err_state:
                logger.debug("Send loop closed as error state is set")
                break
            if self.agent.recieved_pull_the_plug:
                logger.warning("Send loop closed as we recieved the 'Pull the Plug' signal")
                break
            # then listen to the message queue
            message = await self.agent.neosphere_client.get()
            if message is None:
                continue
            # logger.debug(f"Sending message: {message['cmd']}")
            logger.debug(f"Sending message: {message}")
            task = asyncio.create_task(self.ws.send(json.dumps(message)))
            self.send_active.append(task)
        logger.info(f"Send loop ending. Will wait for all outgoing messages ({len(self.send_active)}) to finish.")
        if len(self.send_active) > 0:
            await asyncio.gather(*self.send_active)
        logger.info(f"Send loop ended.")

    async def run(self):
        while self._retry_count <= self.MAX_RETRIES:
            await self.connect()
            await self.authorize()
            await asyncio.gather(self.ws_recv_loop(), self.ws_send_loop())
            # wait for the above coroutines to finish - indicates end
            logger.info("Receiver and Sender loops ended.")
            if self._retry_count >= self.MAX_RETRIES:
                logger.info("Max retries reached. Exiting.")
                break
            else:
                logger.info(f"Retrying ({self._retry_count}/{self.MAX_RETRIES}) connection in {self.GAP_BETWEEN_RETRIES_SEC} seconds.")
                await asyncio.sleep(self.GAP_BETWEEN_RETRIES_SEC)
            self._retry_count = self._retry_count + 1

    def asyncio_run(self):
        try:
            asyncio.run(self.run())

        except KeyboardInterrupt:
            logger.info('Correct exit')



import logging
import requests
import time

from neosphere.media_handler import get_headers
logger = logging.getLogger('neosphere').getChild(__name__)

class SingletonMeta(type):
    """A metaclass for creating singleton classes."""
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]

class NeosphereAgentContactsClient:
    """
    This class converts Niopub agents information into tool schemas for use with function
    calling in LLMs. 
    
    Currently this class supports translation of Niopub agent descriptions to tool schemas for both
    Claude and ChatGPT. The tool schema is a JSON object that describes how your LLM
    can use other agent to solve a specific task and is derived from the Description
    and Input fields of the agent on Niopub app. These fields can be modified on the app
    by the user who owns the agent.

    This allows our local agents to integrate with other online Niopub agents using 
    an LLM's tool calling (aka function calling) capabilities.
    """
    def __init__(self, reconn_token, agent_names, url):
        self.public_cache = {}
        self.private_cache = {}
        self.token = reconn_token
        self.base_url = url
        # self.initial_contacts_setup(agent_names)

    # def initial_contacts_setup(self, agent_names):
        # self._fetch_all_agents(agent_names)
    
    def _fetch_all_agents(self, agent_names):
        """Fetch information for all agents, including owned, and cache it."""
        logger.debug("Fetching all agent contact info")
        self._fetch_and_cache_agents(agent_names, include_owned=True)

    def _fetch_and_cache_agents(self, requested_agent_names, include_owned=False):
        """Fetch information for a list of agents from the API and cache it."""
        payload = {"share_ids": requested_agent_names}
        if include_owned:
            payload["owned"] = True
        logger.debug(f"Fetching agent contact info for (if online and accepting queries): {requested_agent_names}")
        response = requests.post(self.base_url+'post/agent/contacts', json=payload, headers=get_headers(self.token))
        # log response
        if response.status_code == 200:
            data = response.json()
            for agent in data:
                agent_info = {
                    'share_id': agent['share_id'],
                    'description': agent['description'],
                    'status': agent['status'],
                    'owned': False,
                    'ts': time.time()
                }
                if 'input_desc' in agent:
                    agent_info['input_desc'] = agent['input_desc']
                if 'owned' in agent:
                    agent_info['owned'] = True
                if agent.get('open_to_public', False):
                    self.public_cache[agent['share_id']] = agent_info
                else:
                    self.private_cache[agent['share_id']] = agent_info
                return
        else:
            logger.error(f"Failed to fetch contacts with HTTP code: {response.status_code}")
            return
    
    def get_contact_count(self):
        return len(self.public_cache)+len(self.private_cache)

    def add_agent(self, agent_name):
        """Add an agent name to the list and fetch its data."""
        logger.debug(f"Adding agent {agent_name}")
        if not self.get_agent(agent_name):
            self._fetch_and_cache_agents([agent_name])

    def remove_agent(self, agent_name):
        """Remove an agent name from the list and cache."""
        if self.get_agent(agent_name):
            if agent_name in self.public_cache:
                del self.public_cache[agent_name]
            if agent_name in self.private_cache:
                del self.private_cache[agent_name]
    
    def get_agent(self, agent_name, newer_than_sec=300, marked_online=True):
        """Check if an agent exists, is not expired and is online (optional) in the cache."""
        if agent_name in self.public_cache or agent_name in self.private_cache:
            agent = self.public_cache.get(agent_name, self.private_cache.get(agent_name))
            if int(time.time()) - agent.get('ts', 0) > newer_than_sec:
                return None
            if marked_online and agent.get('status', "Offline") == "Offline":
                return None
            return agent
        return None
    
    def get_or_add_agent(self, agent_name) -> bool:
        """Get an agent if it exists, otherwise add it and return the agent data."""
        agent = self.get_agent(agent_name)
        if agent is None:
            self.add_agent(agent_name)
            agent = self.get_agent(agent_name)
        return True if agent else False

    def get_tool_schema(self, agent_names=None, backend=None, only_public_agents=True, only_private_agents=True, self_owned_agents=False):
        """
        Generate tool schemas for agents. These agents can be public, private, or the agents you own. 
        The schema is returned as a list.

        Params:
            agent_names: List of public agent names to fetch. Can be None if we want to fetch only owned or private agents.
            backend: The backend to generate the schema for. Currently supports 'anthropic' and 'openai'.
            only_public_agents: Only add public agents in the resultant schema.
            only_private_agents: Only add private agents in the resultant schema.
            self_owned_agents: Include all agents that you own in the schema.
        
        Returns:
            List of tool schemas for the agents.
        """
        if not only_public_agents and not only_private_agents and not self_owned_agents:
            raise ValueError("At least one of 'only_public_agents' or 'only_private_agents' or 'self_owned_agents' must be True.")
        if backend not in ["anthropic", "openai"]:
            raise ValueError("Unsupported backend specified. Use 'anthropic' or 'openai'.")

        agents_needed = [agent_names] if agent_names else set(self.public_cache.keys()).union(self.private_cache.keys())
        agents_to_fetch = []

        # We can also call get_or_add_agent but this approach is more efficient
        # since it fetches all agents in one go. The get_or_add_agent approach would fetch
        # each agent individually.
        for agent in agents_needed:
            if not self.get_agent(agent):
                agents_to_fetch.append(agent)
        logger.debug(f"Fetching agents for tools schema: {agents_to_fetch}, include_owned: {self_owned_agents}")
        self._fetch_and_cache_agents(agents_to_fetch, include_owned=self_owned_agents)

        response_tool_schemas = []
        for agent in agents_needed:
            if agent in self.public_cache:
                agent_info = self.public_cache[agent]
                if self_owned_agents and agent_info['owned']:
                    response_tool_schemas.append(self._generate_schema(agent_info, backend))
                elif only_public_agents:
                    response_tool_schemas.append(self._generate_schema(agent_info, backend))
            if agent in self.private_cache:
                agent_info = self.private_cache[agent]
                if self_owned_agents and agent_info['owned']:
                    response_tool_schemas.append(self._generate_schema(agent_info, backend))
                elif only_private_agents:
                    response_tool_schemas.append(self._generate_schema(agent_info, backend))
        return response_tool_schemas

    def _generate_schema(self, agent_data, backend):
        """Helper method to generate a tool schema from agent data, following the schema pattern for both Claude and ChatGPT."""
        input_description = agent_data.get('input', {
            "type": "string",
            "description": f"A natural language query requesting the bot to perform its function: {agent_data['description']}"
        })

        if backend == "anthropic":
            schema = {
                "name": agent_data["share_id"],
                "description": agent_data["description"],
                "input_schema": {
                    "type": "object",
                    "properties": {
                        "input": input_description
                    },
                    "required": ["input"]
                }
            }
        elif backend == "openai":
            schema = {
                "type": "function",
                "function": {
                    "name": agent_data["share_id"],
                    "description": agent_data["description"],
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "input": input_description
                        },
                        "required": ["input"],
                        "additionalProperties": False
                    }
                }
            }
        else:
            raise ValueError("Unsupported backend specified. Use 'anthropic' or 'openai'.")
        
        return schema


from requests import HTTPError
from neosphere.contacts_handler import NeosphereAgentContactsClient
from neosphere.media_handler import NeosphereMediaClient
import asyncio
import time
import uuid
import json
from typing import Any, List
import logging
logger = logging.getLogger('neosphere-client').getChild(__name__)


class Message:
    def __init__(self, **kwargs) -> None:
        # check if token in kwargs
        self.token = kwargs.get('token', None)
        self.text: str = kwargs.get('text', None)
        self.data_ids = kwargs.get('data_ids', [])
        self.from_id = kwargs.get('from_id', None)
        self.from_owner = kwargs.get('from_owner', None)
        self.group_id = kwargs.get('group_id', None)
        self.query_id = kwargs.get('query_id', None)
        self.is_resp = kwargs.get('is_resp', False)
        self.is_err = kwargs.get('is_err', False)

    def to_dict(self):
        if self.token:
            return {
                'token': self.token
            }
        return {
            'text': self.text,
            'data_ids': self.data_ids,
            'from_id': self.from_id,
            'from_owner': self.from_owner,
            'group_id': self.group_id,
            'query_id': self.query_id,
            'is_resp': self.is_resp,
            'is_err': self.is_err,
        }

    def to_json(self):
        return json.dumps(self.to_dict())

    def __str__(self):
        return self.to_dict()

    def _compare_text(self, text):
        if self.text and self.text == text:
            return True
        return False

    def is_pull_the_plug(self):
        return self._compare_text('close') and self.from_id == 'sys' and self.group_id == 'sys'

    @staticmethod
    def from_json(json_str):
        message_dict = json.loads(json_str)
        return Message.from_dict(message_dict)

    @staticmethod
    def from_dict(message_dict):
        return Message(**message_dict)


class NeosphereClient(asyncio.Queue):
    """
    Client handles intents that are usually outgoing requests. 
    It then adds the message to the instance's queue if it's a 
    query that needs to be tracked for response.
    """
    def __init__(self, query_index, name, *args, **kwargs) -> None:
        # if media_handler in kwargs then read it into a variable and unset it
        if 'media_handler' in kwargs:
            self.media_handler = kwargs.pop('media_handler')
        else:
            self.media_handler = None
        super().__init__(*args, **kwargs)
        self.query_index = query_index
        self.backoff_signal_template = {'cmd': 'err', 'text': 'w8'}
        self.hold_signal_template = {'cmd': 'err', 'text': 'hold'}
        self.name = name

    async def send(self, item: Any) -> None:
        await self.put(item)

    async def recv(self) -> Any:
        return await self.get()

    def get_query_index(self):
        return self.query_index

    def register_media_handler(self, media_handler: NeosphereMediaClient):
        self.media_handler = media_handler

    def register_contacts_handler(self, contacts: NeosphereAgentContactsClient):
        self.contacts = contacts

    async def get_medias(self, *media_ids):
        media_list = {}
        if len(media_ids) == 1 and isinstance(media_ids[0], list):
            # Called as foo(a, b) where b is a list [b1, b2, b3]
            media_id_list = media_ids[0]
        else:
            # Called as foo(a, b1, b2, b3)
            media_id_list = media_ids
        if self.media_handler:
            for media_id in media_id_list:
                try:
                    x = self.media_handler.get_media(media_id)
                except Exception as e:
                    logger.error(f"Failed to get media {media_id}: {e}")
                    continue
                media_list[media_id] = x
        else:
            logger.error(f"No media handler registered, to get {media_id}.")
        return media_list

    async def create_forward_copy(self, forward_to_id, *media_ids):
        new_list = []
        if len(media_ids) == 1 and isinstance(media_ids[0], list):
            # Called as foo(a, b) where b is a list [b1, b2, b3]
            media_id_list = media_ids[0]
        else:
            # Called as foo(a, b1, b2, b3)
            media_id_list = media_ids
        if self.media_handler:
            for media_id in media_id_list:
                try:
                    new_id = self.media_handler.create_forward_copy_id(forward_to_id, media_id)
                    new_list.append(new_id)
                except HTTPError as e:
                    logger.error(f"Failed to forward media {media_id} to {forward_to_id}.")
                    if e.response.status_code == 401:
                        logger.error(f"Do you have access to {media_id} and is agent {forward_to_id} online?")
        else:
            logger.error(f"No media handler registered, to forward {media_id} to {forward_to_id}.")
        return new_list

    async def save_media(self, parent_id, media_file)->str:
        if self.media_handler:
            return await self.media_handler.save_media(parent_id, media_file)
        else:
            logger.error(f"No media handler registered, to save media.")
            return None

    def save_medias(self, parent_id, media_files):
        """
        Save multiple media files at once.

        Parameters:
            parent_id (str): The ID of the parent media.
            media_files (list): A list of file-like objects containing media data.

        Returns:
            list: A list of media IDs for the saved media.
        """
        media_ids = []
        for media_file in media_files:
            media_id = self.save_media(parent_id, media_file)
            media_ids.append(media_id)
        return media_ids

    async def send_token_request(self, connection_code, agent_share_id, client_name):
        auth_data: dict = {
            'cmd': 'aiagent',
            'code': connection_code, # You get this from the app from your AI Agent's profile
            'id': agent_share_id, # The agent's ID, displayed on Agent profile as niopub.com/x/john.doe
            'client_id': client_name
        }
        await self.send(auth_data)

    async def send_token_to_reconnect(self, token, agent_share_id):
        token_conn = {
            'cmd': 'aiagent',
            'token': token,
            'id': agent_share_id,
        }
        await self.send(token_conn)

    async def respond_to_group_message(self, group_id, response_data, media_ids: List[str]=[], choices: List[str]=[]):
        if not group_id:
            logger.error("Group ID is required to send a response to group message.")
            return None
        group_message = {
            'cmd': 'group-response',
            'group_id': group_id,
            'text': response_data
        }
        if media_ids:
            group_message['data_ids'] = media_ids
        if choices:
            group_message['choices'] = choices
        #put response in send queue
        await self.send(group_message)

    async def query_agent(self, agent_id, query, media_ids: List[str]=[], query_id: str=None):
        if not agent_id:
            logger.error("Agent ID is required to query an agent.")
            return None
        # check if the agent_id is in the contacts and is online
        if not self.contacts.get_or_add_agent(agent_id):
            return None
        # generate a uuid without dashes
        query_id = agent_id + str(uuid.uuid4())[:8] if not query_id else query_id
        query_created = {
            'cmd': 'query',
            'to_id': agent_id,
            'query_id': query_id,
            'text': query
        }
        if media_ids:
            query_created['data_ids'] = media_ids
        # overwrites prev sent record if a query ID is re-used
        self._record_in_query_tracker(query_id, query_created)
        #put query in send queue
        await self.send(query_created)
        return query_id

    def _record_in_query_tracker(self, query_id, query: dict):
        self.query_index[query_id] = {
            'sent_on': int(time.time()),
            # 'text': query
        }
        return query_id

    async def send_backoff_signal(self, to_id):
        """
        If the network sees multiple backoff signals from you to the same agent, 
        we will eventually put the agent on a 30s hold from sending you messages.

        Useful when an agent is sending too many queries and you want to slow them down.
        """
        err_signal = self.backoff_signal_template.copy()
        err_signal['to_id'] = to_id
        await self.send(err_signal)

    async def put_a_30s_hold(self, group_id):
        """
        Putting a 30s hold on a group will prevent any user in the group from 
        sending queries to the agent for 30s.

        Useful when an agent is getting too many queries from a group and you want to slow them down.
        """
        hold_signal = self.hold_signal_template.copy()
        hold_signal['group_id'] = group_id
        await self.send(hold_signal)

    async def _record_query_response_recvd(self, query_id, response: Message):
        """
        This is an internal helper function to record the response received for a query in the query_index.

        This method is intended to be used by the connection handler to record the response received for a query
        and should not be called directly by the client (our AI agent).
        """
        # check if query_id exists in query_index
        if query_id in self.query_index:
            logger.warning(f"Got response for query ID {query_id} (from agent {response.from_id}).")
            # add response to query_index
            self.query_index[query_id]['response_rcv'] = response
        else:
            # create new record
            logger.warning(f"Got response for query ID {query_id} (from agent {response.from_id}). But query is missing from query_index. Dropping the response and sending a lost query signal.")
            await self.send_backoff_signal(response.from_id)
            return

    async def wait_for_query_response(self, query_id, timeout=10, check_interval=0.5) -> Message:
        """
        Given a query_id, this function will wait for the response to the query.

        The wait logic is simple, it checks if the query_id is in the query_index and if it has a response_rcv key.
        The response_rcv key is added to the query_index when a response is received for the query_id on the 
        connection. If the response is not received within the timeout, the function will return None.
        """
        start_time = int(time.time())
        while True:
            logger.debug(f"Checking for query response for query ID: {query_id}...")
            if query_id in self.query_index:
                if 'response_rcv' in self.query_index[query_id]:
                    resp = self.query_index[query_id]['response_rcv']
                    del self.query_index[query_id]
                    return resp
            else:
                # query_id is not in query_index, nothing to wait for
                logger.error(f"Query ID {query_id} not found, nothing to wait for.")
                return None
            if int(time.time()) - start_time > timeout:
                logger.warning(f"Timeout while waiting for query response for query ID: {query_id}")
                return None
            await asyncio.sleep(check_interval)

    async def respond_to_agent_query(self, agent_id, query_id, response_data, media_ids: List[str]=[]):
        """
        Send a response to a query from another agent.

        Requires the query_id you received in the original query to be sent back with this response.
        """
        query_created = {
            'cmd': 'ans',
            'to_id': agent_id,
            'query_id': query_id,
            'text': response_data
        }
        if media_ids:
            query_created['data_ids'] = media_ids
        #put query in send queue
        await self.send(query_created)